#include <sqlextract/filter/filter_comparison.h>
#include <sqlextract/filter/filter_logic.h> /* For filter_create_not */
#include <sqlextract/filter/arbval.h>
#include <sqlextract/util.h>
#include <stdlib.h>
#include <string.h>

struct filter_comparison_storage_t {
	struct filter_t *lhs;
	struct filter_t *rhs;
};

struct filter_comparison_between_storage_t {
	struct filter_t *val;
	struct filter_t *low;
	struct filter_t *high;
	struct arbval_t *tmpresult; /* We don't want to reallocate this all the time */
};

static void filter_comparison_destroy(void *storage) {
	struct filter_comparison_storage_t *stg =
			(struct filter_comparison_storage_t *) storage;

	filter_destroy(stg->lhs);
	filter_destroy(stg->rhs);
	free(stg);
}

/**
 * Test if LHS > RHS
 */
static void filter_process_comp_gt(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_comparison_storage_t *stg =
			(struct filter_comparison_storage_t*) storage;
	struct arbval_t *lhsval = filter_process(stg->lhs, object);
	struct arbval_t *rhsval = filter_process(stg->rhs, object);

	if (lhsval->type == ARBVAL_TYPE_NULL || rhsval->type == ARBVAL_TYPE_NULL) {
		arbval_typecast_null(result);
		return;
	}

	arbval_typecast_integer(result); /* Make sure it can store 0 or 1 */

	switch (lhsval->type) {
	case ARBVAL_TYPE_NULL:
		break;
	case ARBVAL_TYPE_INTEGER:
		if (rhsval->type == ARBVAL_TYPE_INTEGER) {
			result->value.integer = lhsval->value.integer
					> rhsval->value.integer;
			break;
		}
		arbval_typecast_real(lhsval);
		/*no break*/
	case ARBVAL_TYPE_REAL:
		switch (rhsval->type) {
		case ARBVAL_TYPE_NULL:
			break;
		case ARBVAL_TYPE_INTEGER:
			arbval_typecast_real(rhsval);
			/*no break*/
		case ARBVAL_TYPE_REAL:
			result->value.integer = lhsval->value.real > rhsval->value.real;
			break;
		case ARBVAL_TYPE_STRING:
			/* NUM > STR is always true */
			result->value.integer = 1;
			break;
		}
		break;
	case ARBVAL_TYPE_STRING:
		switch (rhsval->type) {
		case ARBVAL_TYPE_NULL:
			break;
		case ARBVAL_TYPE_INTEGER:
		case ARBVAL_TYPE_REAL:
			/* STR > NUM is always false */
			result->value.integer = 0;
			break;
		case ARBVAL_TYPE_STRING:
			result->value.integer = strcmp(lhsval->value.string,
					rhsval->value.string) > 0;
			break;
		}
		break;
	}
}

/**
 * Test if LHS == RHS
 */
static void filter_process_comp_eq(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_comparison_storage_t *stg =
			(struct filter_comparison_storage_t*) storage;
	struct arbval_t *lhsval = filter_process(stg->lhs, object);
	struct arbval_t *rhsval = filter_process(stg->rhs, object);

	if (lhsval->type == ARBVAL_TYPE_NULL || rhsval->type == ARBVAL_TYPE_NULL) {
		arbval_typecast_null(result);
		return;
	}

	arbval_typecast_integer(result); /* Make sure it can store 0 or 1 */

	switch (lhsval->type) {
	case ARBVAL_TYPE_NULL:
		break;
	case ARBVAL_TYPE_INTEGER:
		if (rhsval->type == ARBVAL_TYPE_INTEGER) {
			result->value.integer = lhsval->value.integer
					== rhsval->value.integer;
			break;
		}
		arbval_typecast_real(lhsval);
		/*no break*/
	case ARBVAL_TYPE_REAL:
		switch (rhsval->type) {
		case ARBVAL_TYPE_NULL:
			break;
		case ARBVAL_TYPE_INTEGER:
			arbval_typecast_real(rhsval);
			/*no break*/
		case ARBVAL_TYPE_REAL:
			result->value.integer = lhsval->value.real == rhsval->value.real;
			break;
		case ARBVAL_TYPE_STRING:
			/* NUM == STR is always true */
			result->value.integer = 0;
			break;
		}
		break;
	case ARBVAL_TYPE_STRING:
		switch (rhsval->type) {
		case ARBVAL_TYPE_NULL:
			break;
		case ARBVAL_TYPE_INTEGER:
		case ARBVAL_TYPE_REAL:
			/* STR == NUM is always false */
			result->value.integer = 0;
			break;
		case ARBVAL_TYPE_STRING:
			result->value.integer = strcmp(lhsval->value.string,
					rhsval->value.string) == 0;
			break;
		}
		break;
	}
}

static void filter_process_comp_is(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_comparison_storage_t * stg = storage;
	struct arbval_t *lhsval = filter_process(stg->lhs, object);
	struct arbval_t *rhsval = filter_process(stg->rhs, object);
	/*
	 * The IS and IS NOT operators work like = and != except when one or both of
	 * the operands are NULL. In this case, if both operands are NULL, then the
	 * IS operator evaluates to 1 (true) and the IS NOT operator evaluates to 0
	 * (false). If one operand is NULL and the other is not, then the IS
	 * operator evaluates to 0 (false) and the IS NOT operator is 1 (true). It
	 * is not possible for an IS or IS NOT expression to evaluate to NULL.
	 *
	 */
	arbval_typecast_integer(result);

	if (lhsval->type == ARBVAL_TYPE_NULL && rhsval->type == ARBVAL_TYPE_NULL) {
		result->value.integer = 1;
		return;
	} else if (lhsval->type == ARBVAL_TYPE_NULL
			|| rhsval->type == ARBVAL_TYPE_NULL) {
		result->value.integer = 0;
		return;
	}

	filter_process_comp_eq(result, object, storage);
}

struct filter_t *filter_create_comp_gt(struct filter_t *lhs,
		struct filter_t *rhs) {
	struct filter_comparison_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_comparison_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_comp_gt, filter_comparison_destroy,
			(void*) stg);
}

struct filter_t *filter_create_comp_lt(struct filter_t *lhs,
		struct filter_t *rhs) {
	return filter_create_comp_gt(rhs, lhs);
}

struct filter_t *filter_create_comp_ge(struct filter_t *lhs,
		struct filter_t *rhs) {
	return filter_create_not(filter_create_comp_lt(lhs, rhs));
}

struct filter_t *filter_create_comp_le(struct filter_t *lhs,
		struct filter_t *rhs) {
	return filter_create_not(filter_create_comp_gt(lhs, rhs));
}

struct filter_t *filter_create_comp_eq(struct filter_t *lhs,
		struct filter_t *rhs) {
	struct filter_comparison_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_comparison_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_comp_eq, filter_comparison_destroy,
			(void*) stg);
}

struct filter_t *filter_create_comp_is(struct filter_t *lhs,
		struct filter_t *rhs) {
	struct filter_comparison_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_comparison_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_comp_is, filter_comparison_destroy,
			(void *) stg);
}

static void filter_process_between(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_comparison_between_storage_t *stg =
			(struct filter_comparison_between_storage_t *) storage;
	struct filter_comparison_storage_t lowstg;
	struct filter_comparison_storage_t highstg;

	int lowres;
	int highres;

	/*
	 * We only have filter_process_comp_gt which >
	 *
	 * We model it as:
	 * L <= V AND V <= H
	 * !(!(L <= V) OR !(V <= H))
	 * !( L > V OR V > H)
	 */

	lowstg.lhs = stg->low;
	lowstg.rhs = stg->val;

	highstg.lhs = stg->val;
	highstg.rhs = stg->high;

	filter_process_comp_gt(stg->tmpresult, object, (void*) &lowstg);
	lowres = arbval_is_true(stg->tmpresult);
	filter_process_comp_gt(stg->tmpresult, object, (void*) &highstg);
	highres = arbval_is_true(stg->tmpresult);

	arbval_typecast_integer(result);
	result->value.integer = (lowres || highres) ? 0 : 1;
}

static void filter_between_destroy(void *storage) {
	struct filter_comparison_between_storage_t *stg =
			(struct filter_comparison_between_storage_t *) storage;
	filter_destroy(stg->val);
	filter_destroy(stg->low);
	filter_destroy(stg->high);
	arbval_destroy(stg->tmpresult);
	free(stg);
}

struct filter_t *filter_create_between(struct filter_t *val,
		struct filter_t *low, struct filter_t *high) {
	struct filter_comparison_between_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_comparison_between_storage_t));
	stg->val = val;
	stg->low = low;
	stg->high = high;
	stg->tmpresult = arbval_create_integer(0);
	return filter_create(filter_process_between, filter_between_destroy,
			(void *) stg);
}
