#include <sqlextract/filter/filter_strings.h>
#include <sqlextract/util.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <unicode/uregex.h>
#include <unicode/ustring.h>
#include <assert.h>

struct filter_strings_storage_t {
	struct filter_t *lhs;
	struct filter_t *rhs;
	struct filter_t *escape;
	logfilter_error *error;
	int flags;
};

static void filter_strings_destroy(void *storage) {
	struct filter_strings_storage_t *stg =
		(struct filter_strings_storage_t *) storage;

	filter_destroy(stg->lhs);
	filter_destroy(stg->rhs);
	filter_destroy(stg->escape);
	free(stg);
}

/**
 * The utf8 parts here are ripped from the sqlite3 amalgamation.
 */
static const unsigned char utf8_transl[] = {
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
  0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
  0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
  0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
  0x00, 0x01, 0x02, 0x03, 0x00, 0x01, 0x00, 0x00,
};


static uint32_t utf8_read_char( const unsigned char **pz){
  unsigned int c;
  c = *((*pz)++);
  if( c>=0xc0 ){
    c = utf8_transl[c-0xc0];
    while( (*(*pz) & 0xc0)==0x80 ){
      c = (c<<6) + (0x3f & *((*pz)++));
    }
    if( c<0x80
        || (c&0xFFFFF800)==0xD800
        || (c&0xFFFFFFFE)==0xFFFE ){  c = 0xFFFD; }
  }
  return c;
}

/*
 * pZ is a UTF-8 encoded unicode string. If nByte is less than zero,
 * return the number of unicode characters in pZ up to (but not including)
 * the first 0x00 byte. If nByte is not less than zero, return the
 * number of unicode characters in the first nByte of pZ (or up to
 * the first 0x00, whichever comes first).
 */
static int utf8_char_length(const char *zIn, int nByte){
  int r = 0;
  const uint8_t *z = (const uint8_t*)zIn;
  const uint8_t *zTerm;
  if( nByte>=0 ){
    zTerm = &z[nByte];
  }else{
    zTerm = (const uint8_t*)(-1);
  }

  assert( z<=zTerm );

  while( *z!=0 && z<zTerm ){

	  if( (*(z++)) >= 0xc0 ){
		  /* Skip until next character*/
		  while( (*z & 0xc0)==0x80 ){
			  if( (*(z++))>=0xc0 ){
				  z++;
			  }
		  }
	  }
	  r++;
  }
  return r;
}

/**
 * This is an adaptation of the patternCompare function from the sqlite3 amalgamation.
 * It does not include globbing (which the original does) and does not do case insensitivity
 * for ASCII. The latter because it feels more consistent with the utf8 behaviour.
 *
 * It could potentially be O(n**2). Practically, however, for our use case this won't be
 * a problem, since syslog strings aren't that long.
 */
static int filter_like_match( const uint8_t *zPattern,              /* The pattern */
		const uint8_t *zString,               /* The string to compare against the pattern */
		uint32_t esc                          /* The escape character */
		){
	uint32_t c, c2;
	uint8_t matchOne = '_';
	uint8_t matchAll = '%';

	int prevEscape = 0;     /* True if the previous character was 'escape' */

	while( (c = utf8_read_char(&zPattern))!=0 ){
		if( c==matchAll && !prevEscape ){
			while( (c=utf8_read_char(&zPattern)) == matchAll
					|| c == matchOne ) {
				if( c==matchOne && utf8_read_char(&zString)==0 ) {
					return 0;
				}
			}
			if( c==0 ){
				return 1;
			}else if( c==esc ) {
				c = utf8_read_char(&zPattern);
				if( c==0 ){
					return 0;
				}
			}
			while( (c2 = utf8_read_char(&zString))!=0 ) {
				while( c2 != 0 && c2 != c ){
					c2 = utf8_read_char(&zString);
				}
				if( c2==0 ) return 0;
				if( filter_like_match(zPattern,zString,esc) ) return 1;
			}
			return 0;
		} else if( c==matchOne && !prevEscape ) {
			if( utf8_read_char(&zString)==0 ){
				return 0;
			}
		} else if( esc==c && !prevEscape ) {
			prevEscape = 1;
		} else {
			c2 = utf8_read_char(&zString);
			if( c!=c2 ){
				return 0;
			}
			prevEscape = 0;
		}
	}
	return *zString==0;
}

static int filter_regexp_match(logfilter_error *error, const char *string, const char *pattern, int flags) {
	UErrorCode status = U_ZERO_ERROR;
	URegularExpression *pExpr;
	UBool res;
	/* maybe heap-allocate these based on size of source string? */
	UChar zString[4096];
	UChar zPattern[4096];
	uint32_t regexpFlags;

	if (!string || !pattern) {
		return 0;
	}

	u_strFromUTF8(zString, 4096, NULL, string, -1, &status);
	if ( !U_SUCCESS(status) ){
		*error = ERR_UTF8_CONVERSION_FAILED;
		return 0;
	}

	u_strFromUTF8(zPattern, 4096, NULL, pattern, -1, &status);
	if ( !U_SUCCESS(status) ){
		*error = ERR_UTF8_CONVERSION_FAILED;
		return 0;
	}

	regexpFlags = 0;
	if(flags & FILTER_STRINGS_CASE_INSENSITIVE) {
		regexpFlags |= UREGEX_CASE_INSENSITIVE;
	}
	pExpr = uregex_open(zPattern, -1, regexpFlags, 0, &status);
	if(!U_SUCCESS(status) ){
		assert(!pExpr);
		*error = ERR_INVALID_REGULAR_EXPRESSION;
		return 0;
	}

	/* Configure the text that the regular expression operates on. */
	uregex_setText(pExpr, zString, -1, &status);
	if( !U_SUCCESS(status) ){
		*error = ERR_INTERNAL_ERROR;
		uregex_close(pExpr);
		return 0;
	}

	/* Attempt the match */
	res = uregex_find(pExpr, 0, &status);
	if( !U_SUCCESS(status) ){
		*error = ERR_INTERNAL_ERROR;
		uregex_close(pExpr);
		return 0;
	}

	/* Set the text that the regular expression operates on to a NULL
	 ** pointer. This is not really necessary, but it is tidier than 
	 ** leaving the regular expression object configured with an invalid
	 ** pointer after this function returns.
	 */
	uregex_setText(pExpr, 0, 0, &status);

	uregex_close(pExpr);

	/* Return 1 or 0. */
	if(flags & FILTER_STRINGS_INVERSE)
		return (res ? 0 : 1);
	return (res ? 1 : 0);

}

static void filter_process_regexp(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_strings_storage_t *stg =
		(struct filter_strings_storage_t*) storage;
	struct arbval_t *lhsval = filter_process(stg->lhs, object);
	struct arbval_t *patternval = filter_process(stg->rhs, object);

	if (lhsval->type == ARBVAL_TYPE_NULL || patternval->type == ARBVAL_TYPE_NULL) {
		/* if either side is null, the result is null */
		arbval_typecast_null(result);
		return;
	}

	arbval_typecast_integer(result); /* because bool */
	arbval_typecast_string(lhsval);
	arbval_typecast_string(patternval);

	result->value.integer =
		filter_regexp_match(stg->error, lhsval->value.string, patternval->value.string, stg->flags);
}

static void filter_process_like(struct arbval_t *result, void *object,
		void *storage) {
	struct arbval_t *escapeval = NULL;
	struct filter_strings_storage_t *stg =
		(struct filter_strings_storage_t*) storage;
	struct arbval_t *lhsval = filter_process(stg->lhs, object);
	struct arbval_t *patternval = filter_process(stg->rhs, object);
	const unsigned char *p = NULL;

	uint32_t escape = 0;

	if (lhsval->type == ARBVAL_TYPE_NULL || patternval->type == ARBVAL_TYPE_NULL) {
		/* if either side is null, the result is null */
		arbval_typecast_null(result);
		return;
	}

	arbval_typecast_integer(result);
	arbval_typecast_string(patternval);
	arbval_typecast_string(lhsval);
	if (stg->escape) {
		escapeval = filter_process(stg->escape, object);
		//escapeval = arbval_create_string("X");
		arbval_typecast_string(escapeval);

			/*escape string can only be _one_ UTF-8 encoded character long */
			if (utf8_char_length(escapeval->value.string, -1) != 1) {
				*(stg->error) = ERR_ESCAPE_TOO_LONG;
				result->value.integer = 0;
				return;
			}
			p = (unsigned char *)( escapeval->value.string);
			escape = utf8_read_char(&p);
	}

	result->value.integer = filter_like_match((uint8_t *)patternval->value.string,
			(uint8_t *)lhsval->value.string,
			escape
			);
}

static void filter_process_concat(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_strings_storage_t *stg =
		(struct filter_strings_storage_t *) storage;
	struct arbval_t *lhsval = filter_process(stg->lhs, object);
	struct arbval_t *rhsval = filter_process(stg->rhs, object);
	size_t lhs_len = 0;
	size_t rhs_len = 0;

	if (lhsval->type == ARBVAL_TYPE_NULL || rhsval->type == ARBVAL_TYPE_NULL) {
		arbval_typecast_null(result);
		/* if either side is null, the result is null */
		return;
	}

	arbval_typecast_string(result);
	arbval_typecast_string(lhsval);
	arbval_typecast_string(rhsval);
	/* Are we guaranteed to have 0-terminated buffers, here? I sure hope so ...*/
	lhs_len = strlen(lhsval->value.string);
	rhs_len = strlen(rhsval->value.string);
	free(result->value.string);
	result->value.string = lf_calloc(1, lhs_len + rhs_len + 1);
	strncat(result->value.string, lhsval->value.string, lhs_len); /*move lhs to beginning of buffer */
	strncat(result->value.string+lhs_len, rhsval->value.string, rhs_len); /*... and rhs to end of that*/
}

struct filter_t *filter_create_like(logfilter_error *error, struct filter_t *lhs,
		struct filter_t *pattern, struct filter_t *escape) {
	struct filter_strings_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_strings_storage_t));
	stg->lhs = lhs;
	stg->rhs = pattern;
	stg->escape = escape;
	stg->error = error;
	return filter_create(filter_process_like, filter_strings_destroy,
			(void *) stg);
}

struct filter_t *filter_create_regexp(logfilter_error *error, struct filter_t *lhs,
		struct filter_t *pattern, int flags) {
	struct filter_strings_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_strings_storage_t));
	stg->lhs = lhs;
	stg->rhs = pattern;
	stg->error = error;
	stg->flags = flags;
	return filter_create(filter_process_regexp, filter_strings_destroy,
			(void *) stg);
}

struct filter_t *filter_create_concat(struct filter_t *lhs,
		struct filter_t *rhs) {
	struct filter_strings_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_strings_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_concat, filter_strings_destroy,
			(void *) stg);
}
